package nz.ac.massey.javaecs;
/**
 * ECS manager class.
 * Call this class and its functions to interact correctly with the ECS system.
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 * References:
 * Based on the implementation by Austin Morlan:
 * https://code.austinmorlan.com/austin/ecs - 'A simple C++ Entity Component System' - released under MIT licence
 *  
 */

import java.io.PrintStream;
import java.lang.reflect.Type;
import java.util.BitSet;
/**
 * The ECS manager.
 * <p>
 * See https://git.software.kauripeak.co.nz/BrychanD/JavaECS
 * for documentation and more information.
 */
public class Engine {
    // All internal functions write error messages to errorStream; which defaults to System.err
    // Can be set to a different PrintStream, to allow errors to be logged elsewhere
    protected static PrintStream errorStream = System.err;
    // References to the manager classes
    protected EntityManager entityManager;
    protected ComponentManager componentManager;
    protected SystemManager systemManager;
    /**
     * Initialises the ECS with default values
     * <p>
     * Maximum 1024 enitites default
     */
    public Engine(){
        entityManager = new EntityManager();
        componentManager = new ComponentManager();
        systemManager = new SystemManager();
    }

    /**
     * Initialises the ECS with the specified value
     * @param maxEntities the maximum number of entities to allow
     */
    public Engine(int maxEntities){
        entityManager = new EntityManager(maxEntities);
        componentManager = new ComponentManager();
        systemManager = new SystemManager();
    }

    /**
     * /**
     * Creates a new entity
     * @return the index of the new entity
     * @throws IndexOutOfBoundsException if there are no more entities available
     */
    public Entity createEntity() throws IndexOutOfBoundsException{
        Entity newEntity = entityManager.addEntity();
        if (newEntity == null) throw new IndexOutOfBoundsException("Could not create a new entity");
        return newEntity;
    }

    /**
     * Attempts to resize the maximum number of entities
     * @param newSize the new maximum number of entities
     * @return true if the operation succeeded
     */
    public boolean resizeMaximum(int newSize){
        return entityManager.resize(newSize, systemManager, componentManager);
    }

    /**
     * Signals each manager to remove the specified entity
     * @param entity the entity to destroy
     */
    public void destroyEntity(Entity entity){
        entityManager.removeEntity(entity);
        componentManager.entityDestroyed(entity);
        systemManager.entityDestroyed(entity);
    }

    /**
     * Registers the specified name in the component manager
     * @param type the name to register. Should be the component class name or a suitable name for primitive types
     * @return true if the component type was registered successfully
     */
    public boolean registerComponent(Type type){
        return componentManager.registerComponent(type);
    }

    public Integer getComponentIndex(Type type){
        return componentManager.getComponentIndex(type);
    }

    /**
     * Adds an exisiting component to an exisiting entity
     * @param entity the entity to add the component to
     * @param componentType the class name of the component to add
     * @param component the actual component data
     * @return true if the compnent was added to the entity
     */
    public boolean addComponent(Entity entity, Type componentType, Object component){
        if (entityManager.registerComponent(componentManager.getComponentIndex(componentType), entity))
        {
            systemManager.entityRegistrationsChanged(entity, entityManager.getRegistrations(entity));
            componentManager.addComponentToEntity(componentType, component, entity);
            return true;
        }
        else {
            writeErr("(" + componentType.getTypeName() + ")");
            return false;
        }
    }

    /**
     * Removes the component from the specified entity
     * @param entity the entity to remove the component from
     * @param componentType the class type of the component
     * @return true if the component was removed
     */
    public boolean removeComponent(Entity entity, Type componentType){
        if (entityManager.unregisterComponent(componentManager.getComponentIndex(componentType), entity))
        {
            componentManager.removeComponentFromEntity(componentType, entity);
            systemManager.entityRegistrationsChanged(entity, entityManager.getRegistrations(entity));
            return true;
        }
        else return false;
    }

    /**
     * Gets the actual data of the component associated to the entity.
     * May require casting from Object to the known data type
     * @param entity the entity to retrieve the data for
     * @param componentType the class type of the component
     * @return the component data Object associated with the entity
     */
    public Object getComponentData(Entity entity, Type componentType){
        return componentManager.getComponent(componentType, entity);
    }

    /**
     * Registers the system to 
     * @param systemType
     * @param action
     */
    public boolean registerSystem(Type systemType, ECSSystem action){
        return systemManager.registerSystem(systemType, action);
    }

    /**
     * Sets the specified system's signature to the provided signature
     * @param system the class name of the system to set the signature of
     * @param signature the new signature data
     */
    public void setSystemSignature(Type system, BitSet signature){
        systemManager.setRegistrationSignature(system, signature);
    }
    public Integer getMaxEntities(){
        return entityManager.getMaxSize();
    }

    // Encapsulate syserr writes so they may be redirected out of the lib

    /**
     * Writes an error message to the set PrintStream
     * <p>
     * <i>Default System.err
     * @param message the message to write
     */
    static void writeErr(String message){
        errorStream.println(message);
    }

    /**
     * Writes an object to the set PrintStream
     * <p>
     * <i>Default System.err
     * @param obj the object to write
     */
    static void writeErr(Object obj){
        errorStream.println(obj);
    }

    /**
     * Sets the error PrintStream to the provided PrintStream
     * @param newErr the PrintStream to set as the write destination
     */
    public static void setErr(PrintStream newErr){
        errorStream = newErr;
    }

    /**
     * Gets the current error PrintStream
     * @return the current PrintStream that errors are written to
     */
    public static PrintStream getErr(){
        return errorStream;
    }

    public ComponentManager getComponentManager() {
        return componentManager;
    }

    public EntityManager getEntityManager() {
        return entityManager;
    }

    public SystemManager getSystemManager() {
        return systemManager;
    }

    public int getNumEntities(){
        return entityManager.getNumEntities();
    }
}