package nz.ac.massey.javaecs;
/**
 * System Manager
 * This class manages systems registrations, and keeps a current list
 * of all entities the system should operate on
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 */

import java.lang.reflect.Type;
import java.util.BitSet;
import java.util.Map;
import java.util.HashMap;

class SystemManager{
    //private Map<Type, BitSet> registrationSignatures = new HashMap<>();
    private Map<Type, ECSSystem> systems = new HashMap<>();

    /**
     * Signals the SystemManager that an entity was destroyed. 
     * Removes the entity from each system's tracked entities
     * @param entity the destroyed entity
     */
    protected void entityDestroyed(Entity entity){
        for (Type key : systems.keySet()) {
            systems.get(key).entities.remove(entity);
        }
    }

    /**
     * Signals the SystemManager that an entity had its registrations changed, so
     * evaluate if the entity is still relevant to each system
     * @param entity the entity that was modified
     * @param entityRegistrations the new registrations of the entity
     */
    protected void entityRegistrationsChanged(Entity entity, BitSet entityRegistrations){
        for (Type key : systems.keySet()) {
            // Check if the signature is null
            if (!entityRegistrations.equals(null)){
                BitSet srcCpy = (BitSet)entityRegistrations.clone();
                srcCpy.and(systems.get(key).registrationSet);
                if (srcCpy.equals(systems.get(key).registrationSet)){ // Bitwise check if the entity is subscribed to this system
                    systems.get(key).entities.add(entity);
                }
                else{
                    systems.get(key).entities.remove(entity);
                }
            }
        }
    }

    /**
     * Registers the specified system name and system reference
     * @param systemType the class type of the system
     * @param system the instance of the system
     * @return true if the system was added successfully. False if it was already registered; with an error message written to the log
     */
    protected boolean registerSystem(Type systemType, ECSSystem system){
        if (systems.containsKey(systemType)){
            Engine.writeErr("System \'" + systemType.getTypeName() +  "\' already registered");
            return false;
        }
        systems.put(systemType, system);
        return true;
    }

    /**
     * Sets the registrations the system requires
     * @param systemType the class type of the system
     * @param registrations the BitSet containing the required registrations set to true
     */
    protected void setRegistrationSignature(Type systemType, BitSet registrations){
        systems.get(systemType).registrationSet = registrations;
    }

    
}